from netgen import *
from netgen.meshing import *
import netgen.geom2d as geom2d
import numpy as np
from netgen.meshing import PointId
from functions import *

geo = SplineGeometry()        
geo = MakeCorridor()
ngsmesh = geo.GenerateMesh(maxh=0.05)

from ngsolve import *
mesh = Mesh(ngsmesh)

vertex_to_elements = [ [] for i in mesh.vertices]
for i,el in enumerate(mesh.Elements()):
    for v in el.vertices:
        vertex_to_elements[v.nr].append(i)

npoints = mesh.nv        
ref1 = np.array([0.1,0.1])
ref2 = np.array([2.6,0.45])

point_index1 = np.arange(npoints)
point_index2 = np.arange(npoints)

dist_to_ref1 = np.zeros(npoints)
dist_to_ref2 = np.zeros(npoints)
it=0

pointarray = np.zeros(shape=(2,npoints))

for p in mesh.ngmesh.Points():
    x1,y1,z1 = p.p
    pointarray[0, it] = x1
    pointarray[1, it] = y1
    
    dist_to_ref1[it] = sqrt((x1-ref1[0])**2 + (y1 - ref1[1])**2)
    dist_to_ref2[it] = sqrt((x1-ref2[0])**2 + (y1 - ref2[1])**2)
    it = it+1

sort_index1 = [point_index1 for _,point_index1 in sorted(zip(dist_to_ref1,point_index1))]
sort_index2 = [point_index2 for _,point_index2 in sorted(zip(dist_to_ref2,point_index2))]

fesphi = H1(mesh, order=1, dirichlet="outflow")
gfphi = GridFunction(fesphi)
gfphiold = GridFunction(fesphi)
gfphi.Set(CoefficientFunction(10.0), definedon=mesh.Materials("dummy"))

for i,freedof in enumerate(fesphi.FreeDofs()):
    if not freedof:
        gfphi.vec[i] = 0.0

costs = 1.0
eikonal(gfphi, gfphiold, 10, mesh, vertex_to_elements, sort_index1, sort_index2, pointarray, costs)
Draw(gfphi, mesh, 'phi')
Draw(grad(gfphi), mesh, 'b')

input('have a look at the potential, press any key to continue')

# Parameters
tau_pde = 5e-3
tend = 2

N_pde = tend/tau_pde
N_pde = np.ceil(N_pde)
N_pde = np.int_(N_pde)

time_pde = np.linspace(0,tend, N_pde+1)

d1 = 0.05*0.05
d2 = 0.05*0.05

sd1 = sqrt(d1)
sd2 = sqrt(d2)

D = CoefficientFunction(((d1, 0), (0,d2)), dims=(2,2));

# Start simulation by choosing exact values vmax and rhomax to generate data
vmax = 1.5
v_run = Parameter(vmax) # v_run is changed during the MCMC and therefore has to be a Parameter

alpha = 0.9
beta = 0.95

# In- and outflux
fin =  Parameter(alpha)# finscaled is changed during the MCMC and therefore has to be a Parameter
fout = Parameter(beta)

# define FEM spaces: fes1 - L2 functions living on the elements and fes2 for the facet elements to ensure continuity
order = 2
fes1 = L2(mesh, order=order)
fes2 = FacetFESpace(mesh, order=order) 
fes = FESpace([fes1,fes2])

# Test and trial functions
u,uhat = fes.TrialFunction()
v,vhat = fes.TestFunction()

grad_u = u.Deriv()
grad_v = v.Deriv()
n = specialcf.normal(mesh.dim)
h = specialcf.mesh_size


# Discretization of the Laplace operator - HDG formulation
stab = 10
a = BilinearForm(fes) 
a += SymbolicBFI(D * grad(u) * grad(v))
a += SymbolicBFI((D * grad(u)) * n * (vhat-v)+ (D * grad(v))*n*(uhat-u)+ stab *d1*order*order/h*(u-uhat)*(v-vhat), element_boundary=True)
a.Assemble()

# Transportation vector in x-direction
b = CoefficientFunction( (v_run,0) )
#b = -grad(gfphi)
bn = b*specialcf.normal(2)

# Discretization of the convection operator - DG formulation with upwind
conv = BilinearForm(fes)
conv += SymbolicBFI (-u  * (1-u) * b* grad(v))
conv += SymbolicBFI (bn*IfPos(bn, u * (1-u), u.Other() * (1-u.Other())) * (v-v.Other()), VOL, skeleton=True)
conv += SymbolicBFI (bn*IfPos(bn, fout * u, fin * (1.0-u.Other())) * v, BND, skeleton=True) 

# Mass matrix
m = BilinearForm(fes)
m += SymbolicBFI (u * v)
m.Assemble()

# Solution vector for the MCMC problem
u = GridFunction(fes)
u.components[0].data = 0

# Compute the inverse matrix for the pde time stepping as well as the used constants
hmat = a.mat.CreateMatrix()
hmat.AsVector().data = 0.5 * tau_pde * a.mat.AsVector() + m.mat.AsVector()
inv = hmat.Inverse(fes.FreeDofs())

Draw(u.components[0],mesh,'u')

print("time steps = ", N_pde)
check = FPIMEX4(fes, u, a, m, inv,  conv, tau_pde, N_pde)  

#Draw(u.components[0],mesh,'u')
